@@ -30,7 +30,6 @@ import (
 	"bytes"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"runtime"
@@ -38,13 +37,17 @@ import (
 	"strings"
 	"unicode/utf8"
 
+	"github.com/benhoyt/goawk/internal/compiler"
+	"github.com/benhoyt/goawk/internal/cover"
+	"github.com/benhoyt/goawk/internal/parseutil"
+	"github.com/benhoyt/goawk/internal/resolver"
 	"github.com/benhoyt/goawk/interp"
 	"github.com/benhoyt/goawk/lexer"
 	"github.com/benhoyt/goawk/parser"
 )
 
 const (
-	version    = "v1.20.0"
+	version    = "v1.21.0"
 	copyright  = "GoAWK " + version + " - Copyright (c) 2022 Ben Hoyt"
 	shortUsage = "usage: goawk [-F fs] [-v var=value] [-f progfile | 'prog'] [file ...]"
 	longUsage  = `Standard AWK arguments:
@@ -52,11 +55,7 @@ const (
   -f progfile       load AWK source from progfile (multiple allowed)
   -v var=value      variable assignment (multiple allowed)
 
-Additional GoAWK arguments:
-  -cpuprofile file  write CPU profile to file
-  -d                print parsed syntax tree to stderr (debug mode)
-  -da               print virtual machine assembly instructions to stderr
-  -dt               print variable type information to stderr
+Additional GoAWK features:
   -E progfile       load program, treat as last option, disable var=value args
   -H                parse header row and enable @"field" in CSV input mode
   -h, --help        show this help message
@@ -65,6 +64,16 @@ Additional GoAWK arguments:
   -o mode           use CSV output for print with args (ignore OFS and ORS)
                     'csv|tsv [separator=<char>]'
   -version          show GoAWK version and exit
+
+GoAWK debugging arguments:
+  -coverappend      append to coverage profile instead of overwriting
+  -covermode mode   set coverage mode: set, count (default "set")
+  -coverprofile fn  write coverage profile to file
+  -cpuprofile fn    write CPU profile to file
+  -d                print parsed syntax tree to stdout and exit
+  -da               print VM assembly instructions to stdout and exit
+  -dt               print variable type information to stdout and exit
+  -memprofile fn    write memory profile to file
 `
 )
 
@@ -75,15 +84,18 @@ func main() {
 	var progFiles []string
 	var vars []string
 	fieldSep := " "
-	cpuprofile := ""
+	cpuProfile := ""
 	debug := false
 	debugAsm := false
 	debugTypes := false
-	memprofile := ""
+	memProfile := ""
 	inputMode := ""
 	outputMode := ""
 	header := false
 	noArgVars := false
+	coverMode := cover.ModeUnspecified
+	coverProfile := ""
+	coverAppend := false
 
 	var i int
 argsLoop:
@@ -99,6 +111,20 @@ argsLoop:
 		}
 
 		switch arg {
+		case "-covermode":
+			if i+1 >= len(os.Args) {
+				errorExitf("flag needs an argument: -covermode")
+			}
+			i++
+			coverMode = coverModeFromString(os.Args[i])
+		case "-coverprofile":
+			if i+1 >= len(os.Args) {
+				errorExitf("flag needs an argument: -coverprofile")
+			}
+			i++
+			coverProfile = os.Args[i]
+		case "-coverappend":
+			coverAppend = true
 		case "-E":
 			if i+1 >= len(os.Args) {
 				errorExitf("flag needs an argument: -E")
@@ -131,7 +157,7 @@ argsLoop:
 				errorExitf("flag needs an argument: -cpuprofile")
 			}
 			i++
-			cpuprofile = os.Args[i]
+			cpuProfile = os.Args[i]
 		case "-d":
 			debug = true
 		case "-da":
@@ -154,7 +180,7 @@ argsLoop:
 				errorExitf("flag needs an argument: -memprofile")
 			}
 			i++
-			memprofile = os.Args[i]
+			memProfile = os.Args[i]
 		case "-o":
 			if i+1 >= len(os.Args) {
 				errorExitf("flag needs an argument: -o")
@@ -182,84 +208,109 @@ argsLoop:
 			case strings.HasPrefix(arg, "-v"):
 				vars = append(vars, arg[2:])
 			case strings.HasPrefix(arg, "-cpuprofile="):
-				cpuprofile = arg[12:]
+				cpuProfile = arg[len("-cpuprofile="):]
 			case strings.HasPrefix(arg, "-memprofile="):
-				memprofile = arg[12:]
+				memProfile = arg[len("-memprofile="):]
+			case strings.HasPrefix(arg, "-covermode="):
+				coverMode = coverModeFromString(arg[len("-covermode="):])
+			case strings.HasPrefix(arg, "-coverprofile="):
+				coverProfile = arg[len("-coverprofile="):]
 			default:
 				errorExitf("flag provided but not defined: %s", arg)
 			}
 		}
 	}
+	if coverProfile != "" && coverMode == cover.ModeUnspecified {
+		coverMode = cover.ModeSet
+	}
 
 	// Any remaining args are program and input files
 	args := os.Args[i:]
 
-	var src []byte
-	var stdinBytes []byte // used if there's a parse error
+	fileReader := &parseutil.FileReader{}
 	if len(progFiles) > 0 {
 		// Read source: the concatenation of all source files specified
-		buf := &bytes.Buffer{}
 		progFiles = expandWildcardsOnWindows(progFiles)
 		for _, progFile := range progFiles {
 			if progFile == "-" {
-				b, err := ioutil.ReadAll(os.Stdin)
+				err := fileReader.AddFile("<stdin>", os.Stdin)
 				if err != nil {
 					errorExit(err)
 				}
-				stdinBytes = b
-				_, _ = buf.Write(b)
 			} else {
 				f, err := os.Open(progFile)
 				if err != nil {
 					errorExit(err)
 				}
-				_, err = buf.ReadFrom(f)
+				err = fileReader.AddFile(progFile, f)
 				if err != nil {
 					_ = f.Close()
 					errorExit(err)
 				}
 				_ = f.Close()
 			}
-			// Append newline to file in case it doesn't end with one
-			_ = buf.WriteByte('\n')
 		}
-		src = buf.Bytes()
 	} else {
 		if len(args) < 1 {
 			errorExitf(shortUsage)
 		}
-		src = []byte(args[0])
+		err := fileReader.AddFile("<cmdline>", strings.NewReader(args[0]))
+		if err != nil {
+			errorExit(err)
+		}
 		args = args[1:]
 	}
 
 	// Parse source code and setup interpreter
 	parserConfig := &parser.ParserConfig{
 		DebugTypes:  debugTypes,
-		DebugWriter: os.Stderr,
+		DebugWriter: os.Stdout,
 	}
-	prog, err := parser.ParseProgram(src, parserConfig)
+	prog, err := parser.ParseProgram(fileReader.Source(), parserConfig)
 	if err != nil {
 		if err, ok := err.(*parser.ParseError); ok {
-			name, line := errorFileLine(progFiles, stdinBytes, err.Position.Line)
+			name, line := fileReader.FileLine(err.Position.Line)
 			fmt.Fprintf(os.Stderr, "%s:%d:%d: %s\n",
 				name, line, err.Position.Column, err.Message)
-			showSourceLine(src, err.Position)
+			showSourceLine(fileReader.Source(), err.Position)
 			os.Exit(1)
 		}
 		errorExitf("%s", err)
 	}
 
+	coverage := cover.New(coverMode, coverAppend, fileReader)
+
+	if coverMode != cover.ModeUnspecified {
+		astProgram := &prog.ResolvedProgram.Program
+		coverage.Annotate(astProgram)
+
+		// re-resolve annotated program
+		prog.ResolvedProgram = *resolver.Resolve(astProgram, &resolver.Config{
+			DebugTypes:  parserConfig.DebugTypes,
+			DebugWriter: parserConfig.DebugWriter})
+
+		// re-compile it
+		prog.Compiled, err = compiler.Compile(&prog.ResolvedProgram)
+		if err != nil {
+			errorExitf("%s", err)
+		}
+	}
+
 	if debug {
-		fmt.Fprintln(os.Stderr, prog)
+		fmt.Fprintln(os.Stdout, prog)
 	}
 
 	if debugAsm {
-		err := prog.Disassemble(os.Stderr)
+		err := prog.Disassemble(os.Stdout)
 		if err != nil {
 			errorExitf("could not disassemble program: %v", err)
 		}
 	}
 
+	if debug || debugAsm || debugTypes {
+		os.Exit(0)
+	}
+
 	if header {
 		if inputMode == "" {
 			errorExitf("-H only allowed together with -i")
@@ -300,8 +351,8 @@ argsLoop:
 		config.Vars = append(config.Vars, name, value)
 	}
 
-	if cpuprofile != "" {
-		f, err := os.Create(cpuprofile)
+	if cpuProfile != "" {
+		f, err := os.Create(cpuProfile)
 		if err != nil {
 			errorExitf("could not create CPU profile: %v", err)
 		}
@@ -311,16 +362,25 @@ argsLoop:
 	}
 
 	// Run the program!
-	status, err := interp.ExecProgram(prog, config)
+	interpreter, err := interp.New(prog)
+	status, err := interpreter.Execute(config)
+
 	if err != nil {
 		errorExit(err)
 	}
 
-	if cpuprofile != "" {
+	if coverProfile != "" {
+		err := coverage.WriteProfile(coverProfile, interpreter.Array(cover.ArrayName))
+		if err != nil {
+			errorExitf("unable to write coverage profile: %v", err)
+		}
+	}
+
+	if cpuProfile != "" {
 		pprof.StopCPUProfile()
 	}
-	if memprofile != "" {
-		f, err := os.Create(memprofile)
+	if memProfile != "" {
+		f, err := os.Create(memProfile)
 		if err != nil {
 			errorExitf("could not create memory profile: %v", err)
 		}
@@ -334,6 +394,18 @@ argsLoop:
 	os.Exit(status)
 }
 
+func coverModeFromString(mode string) cover.Mode {
+	switch mode {
+	case "set":
+		return cover.ModeSet
+	case "count":
+		return cover.ModeCount
+	default:
+		errorExitf("-covermode can only be one of: set, count")
+		return cover.ModeUnspecified
+	}
+}
+
 // Show source line and position of error, for example:
 //
 //	BEGIN { x*; }
@@ -347,36 +419,6 @@ func showSourceLine(src []byte, pos lexer.Position) {
 	fmt.Fprintln(os.Stderr, strings.Repeat(" ", runeColumn)+strings.Repeat("   ", numTabs)+"^")
 }
 
-// Determine which filename and line number to display for the overall
-// error line number.
-func errorFileLine(progFiles []string, stdinBytes []byte, errorLine int) (string, int) {
-	if len(progFiles) == 0 {
-		return "<cmdline>", errorLine
-	}
-	startLine := 1
-	for _, progFile := range progFiles {
-		var content []byte
-		if progFile == "-" {
-			progFile = "<stdin>"
-			content = stdinBytes
-		} else {
-			b, err := ioutil.ReadFile(progFile)
-			if err != nil {
-				return "<unknown>", errorLine
-			}
-			content = b
-		}
-		content = append(content, '\n')
-
-		numLines := bytes.Count(content, []byte{'\n'})
-		if errorLine >= startLine && errorLine < startLine+numLines {
-			return progFile, errorLine - startLine + 1
-		}
-		startLine += numLines
-	}
-	return "<unknown>", errorLine
-}
-
 func errorExit(err error) {
 	pathErr, ok := err.(*os.PathError)
 	if ok && os.IsNotExist(err) {
