@@ -448,7 +448,7 @@ func TestCommandLine(t *testing.T) {
 
 func TestDevStdout(t *testing.T) {
 	if runtime.GOOS == "windows" {
-		t.Skip("/dev/stdout not presnt on Windows")
+		t.Skip("/dev/stdout not present on Windows")
 	}
 	runAWKs(t, []string{`BEGIN { print "1"; print "2">"/dev/stdout" }`}, "", "1\n2\n", "")
 }
@@ -588,13 +588,14 @@ func normalizeNewlines(b []byte) []byte {
 	return bytes.Replace(b, []byte("\r\n"), []byte{'\n'}, -1)
 }
 
-func TestInputOutputMode(t *testing.T) {
+func TestGoAWKSpecificOptions(t *testing.T) {
 	tests := []struct {
 		args   []string
 		input  string
 		output string
 		error  string
 	}{
+		// CSV-related options
 		{[]string{"-icsv", "-H", `{ print @"age", @"name" }`}, "name,age\nBob,42\nJane,37", "42 Bob\n37 Jane\n", ""},
 		{[]string{"-i", "csv", "-H", `{ print @"age", @"name" }`}, "name,age\nBob,42\nJane,37", "42 Bob\n37 Jane\n", ""},
 		{[]string{"-icsv", `{ print $2, $1 }`}, "Bob,42\nJane,37", "42 Bob\n37 Jane\n", ""},
@@ -604,6 +605,38 @@ func TestInputOutputMode(t *testing.T) {
 		{[]string{"-iabc", `{}`}, "", "", "invalid input mode \"abc\"\n"},
 		{[]string{"-oxyz", `{}`}, "", "", "invalid output mode \"xyz\"\n"},
 		{[]string{"-H", `{}`}, "", "", "-H only allowed together with -i\n"},
+
+		// Debug options (don't test -dt as its output is not stable)
+		{[]string{"-d", `$1 { print 1+1 }`}, "", `
+$1 {
+    print (1 + 1)
+}
+`[1:], ""},
+		{[]string{"-da", `$1 { print 1+1 }`}, "", `
+        // pattern
+0000    FieldInt 1
+
+        // { body }
+0000    Num 1 (0)
+0002    Num 1 (0)
+0004    Add
+0005    Print 1
+
+`[1:], ""},
+		{[]string{"-d", "-da", `$1 { print 1+1 }`}, "", `
+$1 {
+    print (1 + 1)
+}
+        // pattern
+0000    FieldInt 1
+
+        // { body }
+0000    Num 1 (0)
+0002    Num 1 (0)
+0004    Add
+0005    Print 1
+
+`[1:], ""},
 	}
 
 	for _, test := range tests {
@@ -618,7 +651,7 @@ func TestInputOutputMode(t *testing.T) {
 				}
 			}
 			if stdout != test.output {
-				t.Fatalf("expected %q, got %q", test.output, stdout)
+				t.Fatalf("output differs, got:\n%s\nexpected:\n%s", stdout, test.output)
 			}
 		})
 	}
@@ -652,7 +685,7 @@ age,email,name Sarah 25
 }
 
 func TestCSVDocExamples(t *testing.T) {
-	f, err := os.Open("csv.md")
+	f, err := os.Open("docs/csv.md")
 	if err != nil {
 		t.Fatalf("error opening examples file: %v", err)
 	}
@@ -758,3 +791,160 @@ func TestMandelbrot(t *testing.T) {
 		t.Fatalf("expected:\n%s\ngot:\n%s", expected, stdout)
 	}
 }
+
+func TestCoverPrintAnnotatedSource(t *testing.T) {
+	tests := []struct {
+		sourceFiles        []string
+		mode               string
+		expectedResultFile string
+	}{
+		{[]string{"a1.awk"}, "set", "a1_covermode_set.awk"},
+		{[]string{"a2.awk"}, "count", "a2_covermode_count.awk"},
+		{[]string{"a1.awk", "a2.awk"}, "count", "a1_a2_covermode_count.awk"},
+		{[]string{"a3.awk"}, "set", "a3_covermode_set.awk"},
+	}
+
+	for _, test := range tests {
+		t.Run(test.expectedResultFile, func(t *testing.T) {
+			args := []string{"-d", "-covermode", test.mode}
+			for _, file := range test.sourceFiles {
+				args = append(args, "-f", "testdata/cover/"+file)
+			}
+			stdout, stderr, err := runGoAWK(args, "")
+			if err != nil {
+				t.Fatalf("expected no error, got %v (%q)", err, stderr)
+			}
+			expected, err := ioutil.ReadFile("testdata/cover/" + test.expectedResultFile)
+			if err != nil {
+				t.Fatal(err)
+			}
+			if normalizeAwkSource(stdout) != normalizeAwkSource(string(expected)) {
+				t.Fatalf("output differs, got:\n%s\nexpected:\n%s", stdout, expected)
+			}
+		})
+	}
+}
+
+func normalizeAwkSource(source string) string {
+	p, err := parser.ParseProgram([]byte(source), nil)
+	if err != nil {
+		panic(err)
+	}
+	return p.String()
+}
+
+func TestCoverInvalidArgs(t *testing.T) {
+	tests := []struct {
+		args          []string
+		expectedError string
+	}{
+		{[]string{"-coverprofile"}, "flag needs an argument: -coverprofile"},
+		{[]string{"-covermode"}, "flag needs an argument: -covermode"},
+		{[]string{"-covermode", "wrong"}, "-covermode can only be one of: set, count"},
+		{[]string{"-covermode=wrong"}, "-covermode can only be one of: set, count"},
+	}
+
+	for _, test := range tests {
+		t.Run(test.expectedError, func(t *testing.T) {
+			_, stderr, err := runGoAWK(test.args, "")
+			if err == nil {
+				t.Fatalf("expected error")
+			}
+			stderr = strings.TrimSpace(stderr)
+			if stderr != test.expectedError {
+				t.Fatalf("error differs, got:\n%s\nexpected:\n%s", stderr, test.expectedError)
+			}
+		})
+	}
+}
+
+func TestCover(t *testing.T) {
+	tests := []struct {
+		mode                string
+		coverAppend         bool
+		runs                [][]string
+		expectedCoverReport string
+	}{
+		{"set", true, [][]string{{"a1.awk"}}, "test_set.cov"},
+		{"count", true, [][]string{{"a1.awk"}}, "test_count.cov"},
+		{"set", true, [][]string{{"a2.awk", "a1.awk"}}, "test_a2a1_set.cov"},
+		{"count", true, [][]string{{"a2.awk", "a1.awk"}}, "test_a2a1_count.cov"},
+		{"set", true, [][]string{{"a1.awk"}, {"a1.awk"}}, "test_1file2runs_set.cov"},
+		{"count", true, [][]string{{"a2.awk", "a1.awk"}, {"a2.awk", "a1.awk"}}, "test_2file2runs_count.cov"},
+		{"set", false, [][]string{{"a1.awk"}, {"a1.awk"}}, "test_1file2runs_set_truncated.cov"},
+		{"count", false, [][]string{{"a2.awk", "a1.awk"}, {"a2.awk", "a1.awk"}}, "test_2file2runs_count_truncated.cov"},
+	}
+
+	for _, test := range tests {
+		t.Run(test.expectedCoverReport, func(t *testing.T) {
+			tempFile, err := ioutil.TempFile("", "testCov*.txt")
+			if err != nil {
+				t.Fatalf("%v", err)
+			}
+			err = tempFile.Close()
+			if err != nil {
+				t.Fatalf("%v", err)
+			}
+			coverProfile := tempFile.Name()
+			defer os.Remove(coverProfile)
+
+			// make sure file doesn't exist - delete the temp file first, but use its name
+			err = os.Remove(coverProfile)
+			if err != nil {
+				t.Fatalf("%v", err)
+			}
+
+			for _, run := range test.runs {
+				var args []string
+				for _, file := range run {
+					args = append(args, "-f", "testdata/cover/"+file)
+				}
+				args = append(args, "-coverprofile", coverProfile)
+				args = append(args, "-covermode", test.mode)
+				if test.coverAppend {
+					args = append(args, "-coverappend")
+				}
+				_, _, err := runGoAWK(args, "")
+				if err != nil {
+					t.Fatalf("%v", err)
+				}
+			}
+
+			result, err := ioutil.ReadFile(coverProfile)
+			if err != nil {
+				t.Fatalf("%v", err)
+			}
+			resultStr := string(normalizeNewlines(result))
+			resultStr = strings.TrimSpace(convertPathsToFilenames(t, resultStr))
+			expected, err := ioutil.ReadFile("testdata/cover/" + test.expectedCoverReport)
+			if err != nil {
+				t.Fatalf("%v", err)
+			}
+			expectedStr := strings.TrimSpace(string(normalizeNewlines(expected)))
+			if resultStr != expectedStr {
+				t.Fatalf("wrong coverage report, expected:\n\n%s\n\nactual:\n\n%s", expectedStr, resultStr)
+			}
+		})
+	}
+}
+
+func convertPathsToFilenames(t *testing.T, str string) string {
+	lines := strings.Split(str, "\n")
+	for i, line := range lines {
+		if i == 0 {
+			continue // skip mode line
+		}
+		if line == "" {
+			continue // skip empty line
+		}
+		parts := strings.Split(line, ":") // on win line can be `D:\a\goawk\goawk\testdata\cover\a1.awk:2.3,6.30 5 1`
+		path := strings.Join(parts[:len(parts)-1], ":")
+		data := parts[len(parts)-1]
+		if !filepath.IsAbs(path) {
+			t.Fatalf("must be absolute path in coverage report: %s", line)
+		}
+		// leave only the part with name
+		lines[i] = filepath.Base(path) + ":" + data
+	}
+	return strings.Join(lines, "\n")
+}
