@@ -13,11 +13,17 @@ import (
 // Program is an entire AWK program.
 type Program struct {
 	Begin     []Stmts
-	Actions   []Action
+	Actions   []*Action
 	End       []Stmts
-	Functions []Function
-	Scalars   map[string]int
-	Arrays    map[string]int
+	Functions []*Function
+}
+
+// ResolvedProgram is a parsed AWK program + additional data prepared by resolve step
+// needed for subsequent interpretation
+type ResolvedProgram struct {
+	Program
+	Scalars map[string]int
+	Arrays  map[string]int
 }
 
 // String returns an indented, pretty-printed version of the parsed
@@ -75,8 +81,54 @@ func (a *Action) String() string {
 	return strings.Join(patterns, ", ") + sep + stmtsStr
 }
 
+// Node is an interface to be satisfied by all AST elements.
+// We need it to be able to work with AST in a generic way, like in ast.Walk().
+type Node interface {
+	node()
+}
+
+// All these types implement the Node interface.
+func (p *Program) node()        {}
+func (a *Action) node()         {}
+func (f *Function) node()       {}
+func (e *FieldExpr) node()      {}
+func (e *NamedFieldExpr) node() {}
+func (e *UnaryExpr) node()      {}
+func (e *BinaryExpr) node()     {}
+func (e *ArrayExpr) node()      {}
+func (e *InExpr) node()         {}
+func (e *CondExpr) node()       {}
+func (e *NumExpr) node()        {}
+func (e *StrExpr) node()        {}
+func (e *RegExpr) node()        {}
+func (e *VarExpr) node()        {}
+func (e *IndexExpr) node()      {}
+func (e *AssignExpr) node()     {}
+func (e *AugAssignExpr) node()  {}
+func (e *IncrExpr) node()       {}
+func (e *CallExpr) node()       {}
+func (e *UserCallExpr) node()   {}
+func (e *MultiExpr) node()      {}
+func (e *GetlineExpr) node()    {}
+func (s *PrintStmt) node()      {}
+func (s *PrintfStmt) node()     {}
+func (s *ExprStmt) node()       {}
+func (s *IfStmt) node()         {}
+func (s *ForStmt) node()        {}
+func (s *ForInStmt) node()      {}
+func (s *WhileStmt) node()      {}
+func (s *DoWhileStmt) node()    {}
+func (s *BreakStmt) node()      {}
+func (s *ContinueStmt) node()   {}
+func (s *NextStmt) node()       {}
+func (s *ExitStmt) node()       {}
+func (s *DeleteStmt) node()     {}
+func (s *ReturnStmt) node()     {}
+func (s *BlockStmt) node()      {}
+
 // Expr is the abstract syntax tree for any AWK expression.
 type Expr interface {
+	Node
 	expr()
 	String() string
 }
@@ -154,6 +206,7 @@ type ArrayExpr struct {
 	Scope VarScope
 	Index int
 	Name  string
+	Pos   Position
 }
 
 func (e *ArrayExpr) String() string {
@@ -221,6 +274,9 @@ func (e *RegExpr) String() string {
 	return "/" + escaped + "/"
 }
 
+// meaning it will be set during resolve step
+const resolvedLater = -1
+
 type VarScope int
 
 const (
@@ -236,6 +292,7 @@ type VarExpr struct {
 	Scope VarScope
 	Index int
 	Name  string
+	Pos   Position
 }
 
 func (e *VarExpr) String() string {
@@ -315,6 +372,7 @@ type UserCallExpr struct {
 	Index  int
 	Name   string
 	Args   []Expr
+	Pos    Position
 }
 
 func (e *UserCallExpr) String() string {
@@ -375,6 +433,9 @@ func IsLValue(expr Expr) bool {
 
 // Stmt is the abstract syntax tree for any AWK statement.
 type Stmt interface {
+	Node
+	StartPos() Position // position of first character belonging to the node
+	EndPos() Position   // position of first character immediately after the node
 	stmt()
 	String() string
 }
@@ -396,11 +457,45 @@ func (s *DeleteStmt) stmt()   {}
 func (s *ReturnStmt) stmt()   {}
 func (s *BlockStmt) stmt()    {}
 
+func (s *PrintStmt) StartPos() Position    { return s.Start }
+func (s *PrintfStmt) StartPos() Position   { return s.Start }
+func (s *ExprStmt) StartPos() Position     { return s.Start }
+func (s *IfStmt) StartPos() Position       { return s.Start }
+func (s *ForStmt) StartPos() Position      { return s.Start }
+func (s *ForInStmt) StartPos() Position    { return s.Start }
+func (s *WhileStmt) StartPos() Position    { return s.Start }
+func (s *DoWhileStmt) StartPos() Position  { return s.Start }
+func (s *BreakStmt) StartPos() Position    { return s.Start }
+func (s *ContinueStmt) StartPos() Position { return s.Start }
+func (s *NextStmt) StartPos() Position     { return s.Start }
+func (s *ExitStmt) StartPos() Position     { return s.Start }
+func (s *DeleteStmt) StartPos() Position   { return s.Start }
+func (s *ReturnStmt) StartPos() Position   { return s.Start }
+func (s *BlockStmt) StartPos() Position    { return s.Start }
+
+func (s *PrintStmt) EndPos() Position    { return s.End }
+func (s *PrintfStmt) EndPos() Position   { return s.End }
+func (s *ExprStmt) EndPos() Position     { return s.End }
+func (s *IfStmt) EndPos() Position       { return s.End }
+func (s *ForStmt) EndPos() Position      { return s.End }
+func (s *ForInStmt) EndPos() Position    { return s.End }
+func (s *WhileStmt) EndPos() Position    { return s.End }
+func (s *DoWhileStmt) EndPos() Position  { return s.End }
+func (s *BreakStmt) EndPos() Position    { return s.End }
+func (s *ContinueStmt) EndPos() Position { return s.End }
+func (s *NextStmt) EndPos() Position     { return s.End }
+func (s *ExitStmt) EndPos() Position     { return s.End }
+func (s *DeleteStmt) EndPos() Position   { return s.End }
+func (s *ReturnStmt) EndPos() Position   { return s.End }
+func (s *BlockStmt) EndPos() Position    { return s.End }
+
 // PrintStmt is a statement like print $1, $3.
 type PrintStmt struct {
 	Args     []Expr
 	Redirect Token
 	Dest     Expr
+	Start    Position
+	End      Position
 }
 
 func (s *PrintStmt) String() string {
@@ -424,6 +519,8 @@ type PrintfStmt struct {
 	Args     []Expr
 	Redirect Token
 	Dest     Expr
+	Start    Position
+	End      Position
 }
 
 func (s *PrintfStmt) String() string {
@@ -432,7 +529,9 @@ func (s *PrintfStmt) String() string {
 
 // ExprStmt is statement like a bare function call: my_func(x).
 type ExprStmt struct {
-	Expr Expr
+	Expr  Expr
+	Start Position
+	End   Position
 }
 
 func (s *ExprStmt) String() string {
@@ -441,9 +540,12 @@ func (s *ExprStmt) String() string {
 
 // IfStmt is an if or if-else statement.
 type IfStmt struct {
-	Cond Expr
-	Body Stmts
-	Else Stmts
+	Cond      Expr
+	BodyStart Position
+	Body      Stmts
+	Else      Stmts
+	Start     Position
+	End       Position
 }
 
 func (s *IfStmt) String() string {
@@ -456,10 +558,13 @@ func (s *IfStmt) String() string {
 
 // ForStmt is a C-like for loop: for (i=0; i<10; i++) print i.
 type ForStmt struct {
-	Pre  Stmt
-	Cond Expr
-	Post Stmt
-	Body Stmts
+	Pre       Stmt
+	Cond      Expr
+	Post      Stmt
+	BodyStart Position
+	Body      Stmts
+	Start     Position
+	End       Position
 }
 
 func (s *ForStmt) String() string {
@@ -480,9 +585,12 @@ func (s *ForStmt) String() string {
 
 // ForInStmt is a for loop like for (k in a) print k, a[k].
 type ForInStmt struct {
-	Var   *VarExpr
-	Array *ArrayExpr
-	Body  Stmts
+	Var       *VarExpr
+	Array     *ArrayExpr
+	BodyStart Position
+	Body      Stmts
+	Start     Position
+	End       Position
 }
 
 func (s *ForInStmt) String() string {
@@ -491,8 +599,11 @@ func (s *ForInStmt) String() string {
 
 // WhileStmt is a while loop.
 type WhileStmt struct {
-	Cond Expr
-	Body Stmts
+	Cond      Expr
+	BodyStart Position
+	Body      Stmts
+	Start     Position
+	End       Position
 }
 
 func (s *WhileStmt) String() string {
@@ -501,8 +612,10 @@ func (s *WhileStmt) String() string {
 
 // DoWhileStmt is a do-while loop.
 type DoWhileStmt struct {
-	Body Stmts
-	Cond Expr
+	Body  Stmts
+	Cond  Expr
+	Start Position
+	End   Position
 }
 
 func (s *DoWhileStmt) String() string {
@@ -510,21 +623,30 @@ func (s *DoWhileStmt) String() string {
 }
 
 // BreakStmt is a break statement.
-type BreakStmt struct{}
+type BreakStmt struct {
+	Start Position
+	End   Position
+}
 
 func (s *BreakStmt) String() string {
 	return "break"
 }
 
 // ContinueStmt is a continue statement.
-type ContinueStmt struct{}
+type ContinueStmt struct {
+	Start Position
+	End   Position
+}
 
 func (s *ContinueStmt) String() string {
 	return "continue"
 }
 
 // NextStmt is a next statement.
-type NextStmt struct{}
+type NextStmt struct {
+	Start Position
+	End   Position
+}
 
 func (s *NextStmt) String() string {
 	return "next"
@@ -533,6 +655,8 @@ func (s *NextStmt) String() string {
 // ExitStmt is an exit statement.
 type ExitStmt struct {
 	Status Expr
+	Start  Position
+	End    Position
 }
 
 func (s *ExitStmt) String() string {
@@ -547,6 +671,8 @@ func (s *ExitStmt) String() string {
 type DeleteStmt struct {
 	Array *ArrayExpr
 	Index []Expr
+	Start Position
+	End   Position
 }
 
 func (s *DeleteStmt) String() string {
@@ -560,6 +686,8 @@ func (s *DeleteStmt) String() string {
 // ReturnStmt is a return statement.
 type ReturnStmt struct {
 	Value Expr
+	Start Position
+	End   Position
 }
 
 func (s *ReturnStmt) String() string {
@@ -572,7 +700,9 @@ func (s *ReturnStmt) String() string {
 
 // BlockStmt is a stand-alone block like { print "x" }.
 type BlockStmt struct {
-	Body Stmts
+	Body  Stmts
+	Start Position
+	End   Position
 }
 
 func (s *BlockStmt) String() string {
@@ -585,6 +715,7 @@ type Function struct {
 	Params []string
 	Arrays []bool
 	Body   Stmts
+	Pos    Position
 }
 
 func (f *Function) String() string {
@@ -598,3 +729,38 @@ func trimParens(s string) string {
 	}
 	return s
 }
+
+// VarRef is a constructor for *VarExpr
+func VarRef(name string, pos Position) *VarExpr {
+	return &VarExpr{resolvedLater, resolvedLater, name, pos}
+}
+
+// ArrayRef is a constructor for *ArrayExpr
+func ArrayRef(name string, pos Position) *ArrayExpr {
+	return &ArrayExpr{resolvedLater, resolvedLater, name, pos}
+}
+
+// UserCall is a constructor for *UserCallExpr
+func UserCall(name string, args []Expr, pos Position) *UserCallExpr {
+	return &UserCallExpr{false, resolvedLater, name, args, pos}
+}
+
+// PositionError represents an error bound to specific position in source.
+type PositionError struct {
+	// Source line/column position where the error occurred.
+	Position Position
+	// Error message.
+	Message string
+}
+
+// PosErrorf like fmt.Errorf, but with an explicit position.
+func PosErrorf(pos Position, format string, args ...interface{}) error {
+	message := fmt.Sprintf(format, args...)
+	return &PositionError{pos, message}
+}
+
+// Error returns a formatted version of the error, including the line
+// and column numbers.
+func (e *PositionError) Error() string {
+	return fmt.Sprintf("parse error at %d:%d: %s", e.Position.Line, e.Position.Column, e.Message)
+}
